@@ -1,23 +1,22 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright (C) 2008 John Paulett (john -at- paulett.org)
-# Copyright (C) 2009, 2011, 2013 David Aguilar (davvid -at- gmail.com)
+# Copyright (C) 2009, 2011, 2013 David Aguilar (davvid -at- gmail.com) and contributors
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
-
+from __future__ import absolute_import, division, unicode_literals
 import base64
 import warnings
 import sys
 from itertools import chain, islice
 
-import jsonpickle.util as util
-import jsonpickle.tags as tags
-import jsonpickle.handlers as handlers
-
-from jsonpickle.backend import JSONBackend
-from jsonpickle.compat import numeric_types, unicode, PY3, PY2
+from . import util
+from . import tags
+from . import handlers
+from .backend import JSONBackend
+from .compat import numeric_types, unicode, PY3, PY2
 
 
 def encode(value,
@@ -284,48 +283,79 @@ class Pickler(object):
             return handler(self).flatten(obj, data)
 
         reduce_val = None
-        if has_class and not util.is_module(obj):
-            if self.unpicklable:
-                class_name = util.importable_name(cls)
-                data[tags.OBJECT] = class_name
 
-            # test for a reduce implementation, and redirect before doing anything else
-            # if that is what reduce requests
-            if has_reduce_ex:
+        if self.unpicklable:
+            if has_reduce and not has_reduce_ex:
+                try:
+                    reduce_val = obj.__reduce__()
+                except TypeError:
+                    # A lot of builtin types have a reduce which
+                    # just raises a TypeError
+                    # we ignore those
+                    pass
+
+            # test for a reduce implementation, and redirect before
+            # doing anything else if that is what reduce requests
+            elif has_reduce_ex:
                 try:
                     # we're implementing protocol 2
                     reduce_val = obj.__reduce_ex__(2)
                 except TypeError:
-                    # A lot of builtin types have a reduce which just raises a TypeError
+                    # A lot of builtin types have a reduce which
+                    # just raises a TypeError
                     # we ignore those
                     pass
 
-            if has_reduce and not reduce_val:
+            if reduce_val and isinstance(reduce_val, (str, unicode)):
                 try:
-                    reduce_val = obj.__reduce__()
-                except TypeError:
-                    # A lot of builtin types have a reduce which just raises a TypeError
-                    # we ignore those
-                    pass
-
-            if reduce_val:
-                try:
-                    # At this stage, we only handle the case where __reduce__ returns a string
-                    # other reduce functionality is implemented further down
-                    if isinstance(reduce_val, (str, unicode)):
-                        varpath = iter(reduce_val.split('.'))
-                        # curmod will be transformed by the loop into the value to pickle
-                        curmod = sys.modules[next(varpath)]
-                        for modname in varpath:
-                            curmod = getattr(curmod, modname)
-                            # replace obj with value retrieved
+                    varpath = iter(reduce_val.split('.'))
+                    # curmod will be transformed by the
+                    # loop into the value to pickle
+                    curmod = sys.modules[next(varpath)]
+                    for modname in varpath:
+                        curmod = getattr(curmod, modname)
+                        # replace obj with value retrieved
                         return self._flatten(curmod)
                 except KeyError:
                     # well, we can't do anything with that, so we ignore it
                     pass
 
+            elif reduce_val:
+                # at this point, reduce_val should be some kind of iterable
+                # pad out to len 5
+                rv_as_list = list(reduce_val)
+                insufficiency = 5 - len(rv_as_list)
+                if insufficiency:
+                    rv_as_list += [None] * insufficiency
+
+                if rv_as_list[0].__name__ == '__newobj__':
+                    rv_as_list[0] = tags.NEWOBJ
+
+                f, args, state, listitems, dictitems = rv_as_list
+
+                # check that getstate/setstate is sane
+                if not (state and hasattr(obj, '__getstate__')
+                            and not hasattr(obj, '__setstate__')
+                            and not isinstance(obj, dict)):
+                    # turn iterators to iterables for convenient serialization
+                    if rv_as_list[3]:
+                        rv_as_list[3] = tuple(rv_as_list[3])
+
+                    if rv_as_list[4]:
+                        rv_as_list[4] = tuple(rv_as_list[4])
+
+                    data[tags.REDUCE] = list(map(self._flatten, rv_as_list))
+
+                    return data
+
+        if has_class and not util.is_module(obj):
+            if self.unpicklable:
+                class_name = util.importable_name(cls)
+                data[tags.OBJECT] = class_name
+
             if has_getnewargs_ex:
-                data[tags.NEWARGSEX] = list(map(self._flatten, obj.__getnewargs_ex__()))
+                data[tags.NEWARGSEX] = list(
+                    map(self._flatten, obj.__getnewargs_ex__()))
 
             if has_getnewargs and not has_getnewargs_ex:
                 data[tags.NEWARGS] = self._flatten(obj.__getnewargs__())
@@ -359,37 +389,11 @@ class Pickler(object):
         if util.is_sequence_subclass(obj):
             return self._flatten_sequence_obj(obj, data)
 
-        if util.is_noncomplex(obj):
-            return [self._flatten(v) for v in obj]
-
         if util.is_iterator(obj):
             # force list in python 3
             data[tags.ITERATOR] = list(map(self._flatten, islice(obj, self._max_iter)))
             return data
 
-        if reduce_val and not isinstance(reduce_val, (str, unicode)):
-            # at this point, reduce_val should be some kind of iterable
-            # pad out to len 5
-            rv_as_list = list(reduce_val)
-            insufficiency = 5 - len(rv_as_list)
-            if insufficiency:
-                rv_as_list += [None] * insufficiency
-
-            if rv_as_list[0].__name__ == '__newobj__':
-                rv_as_list[0] = tags.NEWOBJ
-
-            data[tags.REDUCE] = list(map(self._flatten, rv_as_list))
-
-            # lift out iterators, so we don't have to iterator and uniterator their content
-            # on unpickle
-            if data[tags.REDUCE][3]:
-                data[tags.REDUCE][3] = data[tags.REDUCE][3][tags.ITERATOR]
-
-            if data[tags.REDUCE][4]:
-                data[tags.REDUCE][4] = data[tags.REDUCE][4][tags.ITERATOR]
-
-            return data
-
         if has_dict:
             # Support objects that subclasses list and set
             if util.is_sequence_subclass(obj):
@@ -532,8 +536,8 @@ class Pickler(object):
 def _mktyperef(obj):
     """Return a typeref dictionary
 
-    >>> _mktyperef(AssertionError)
-    {'py/type': '__builtin__.AssertionError'}
+    >>> _mktyperef(AssertionError) == {'py/type': '__builtin__.AssertionError'}
+    True
 
     """
     return {tags.TYPE: util.importable_name(obj)}
